@@ -9,75 +9,29 @@
 #include <signal.h>
 #include <errno.h>
 #include <unistd.h>
+#include <string.h>
 #include <stdint.h>
 #include <libgen.h>
+#include <signal.h>
 #include <sys/ioctl.h>
 #include <sys/socket.h>
 #include <sys/un.h>
 #include <sys/uio.h>
 #include <sys/poll.h>
+#include <sys/types.h>
+#include <sys/wait.h>
+#include <pwd.h>
 
 #include <vde.h>
+#include <libvdeplug/libvdeplug.h>
 
 #define SWITCH_MAGIC 0xfeedface
 #define BUFSIZE 2048
 #define ETH_ALEN 6
+#define MAXDESCR 128
 
-enum request_type { REQ_NEW_CONTROL };
-
-struct request_v3 {
-  uint32_t magic;
-  uint32_t version;
-  enum request_type type;
-  struct sockaddr_un sock;
-};
-
-
-static int send_fd(char *name, int fddata, struct sockaddr_un *datasock, int intno, int group)
-{
-  int pid = getpid();
-  struct request_v3 req;
-  int fdctl;
-
-  struct sockaddr_un sock;
-
-  if((fdctl = socket(AF_UNIX, SOCK_STREAM, 0)) < 0){
-    perror("socket");
-    exit(1);
-  }
-
-  sock.sun_family = AF_UNIX;
-  snprintf(sock.sun_path, sizeof(sock.sun_path), "%s", name);
-  if(connect(fdctl, (struct sockaddr *) &sock, sizeof(sock))){
-    perror("connect");
-    exit(1);
-  }
-
-  req.magic=SWITCH_MAGIC;
-  req.version=3;
-  req.type=REQ_NEW_CONTROL+((group > 0)?((geteuid()<<8) + group) << 8:0);
-  
-  req.sock.sun_family=AF_UNIX;
-  memset(req.sock.sun_path, 0, sizeof(req.sock.sun_path));
-  sprintf(&req.sock.sun_path[1], "%5d-%2d", pid, intno);
-
-  if(bind(fddata, (struct sockaddr *) &req.sock, sizeof(req.sock)) < 0){
-    perror("bind");
-    exit(1);
-  }
-
-  if (send(fdctl,&req,sizeof(req),0) < 0) {
-    perror("send");
-    exit(1);
-  }
-
-  if (recv(fdctl,datasock,sizeof(struct sockaddr_un),0)<0) {
-	  perror("recv");
-	  exit(1);
-  }
-
-  return fdctl;
-}
+static int nb_nics;
+VDECONN **conn;
 
 unsigned char bufin[BUFSIZE];
 
@@ -115,31 +69,91 @@ static int countnewnics(int argc,char *argv[])
 	return nics;
 }
 
-static void usage() 
+static void usage(void) 
 {
-	if (strcmp(vdeqname,"vdeq") != 0 && strncmp(vdeqname,"vde",3)==0) 
+	if (strcmp(vdeqname,"vdeq") != 0 && strncmp(vdeqname,"vde",3)==0) {
 		fprintf(stderr,"Usage: %s [-h]\n"
 				"\t %s ...qemu options... -net vde[,vlan=n][,sock=sock] ... \n"
 				"Old syntax:\n"
 				"\t %s  [-sock sock1 [,sock2...]] qemu_options\n"
-				"\t (%s executes a qemu machine named %s)\n", vdeqname, vdeqname,vdeqname,vdeqname,filename);
-	else 
+				"\t (%s executes a qemu machine named %s, \n\t  output of \"%s -h\" follows)\n\n", vdeqname,vdeqname,vdeqname,vdeqname,filename,filename);
+		execlp(filename,filename,"-h",(char *) 0);
+	} else {
 		fprintf(stderr,"Usage: %s [-h]\n"
 				"\t %s qemu_executable ...qemu options... -net vde[,vlan=n][,sock=sock] ... \n"
 				"Old syntax:\n"
-				"\t %s qemu_executable [-sock sock1 [,sock2...]] qemu_options\n", vdeqname, vdeqname, vdeqname);
-	exit(0);
+				"\t %s qemu_executable [-sock sock1 [,sock2...]] qemu_options\n", vdeqname,vdeqname, vdeqname);
+		exit(0);
+	}
 }
 
-static void leave()
+static void cleanup()
+{
+	register int i;
+	for (i=0; i<nb_nics; i++) {
+		if (conn[i] != NULL)
+			vde_close(conn[i]);
+	}
+}
+
+static void sig_handler(int sig)
+{
+	fprintf(stderr,"%s: Caught signal %d, cleaning up and exiting\n", vdeqname, sig);
+	cleanup();
+	signal(sig, SIG_DFL);
+	kill(getpid(), sig);
+}
+
+static void setsighandlers()
+{
+	/* setting signal handlers.
+	 * sets clean termination for SIGHUP, SIGINT and SIGTERM, and simply
+	 * ignores all the others signals which could cause termination. */
+	struct { int sig; const char *name; int ignore; } signals[] = {
+		{ SIGHUP, "SIGHUP", 0 },
+		{ SIGINT, "SIGINT", 0 },
+		{ SIGPIPE, "SIGPIPE", 1 },
+		{ SIGALRM, "SIGALRM", 1 },
+		{ SIGTERM, "SIGTERM", 0 },
+		{ SIGUSR1, "SIGUSR1", 1 },
+		{ SIGUSR2, "SIGUSR2", 1 },
+		{ SIGPROF, "SIGPROF", 1 },
+		{ SIGVTALRM, "SIGVTALRM", 1 },
+#ifdef VDE_LINUX
+		{ SIGPOLL, "SIGPOLL", 1 },
+#ifdef SIGSTKFLT
+		{ SIGSTKFLT, "SIGSTKFLT", 1 },
+#endif
+		{ SIGIO, "SIGIO", 1 },
+		{ SIGPWR, "SIGPWR", 1 },
+#ifdef SIGUNUSED
+		{ SIGUNUSED, "SIGUNUSED", 1 },
+#endif
+#endif
+#ifdef VDE_DARWIN
+		{ SIGXCPU, "SIGXCPU", 1 },
+		{ SIGXFSZ, "SIGXFSZ", 1 },
+#endif
+		{ 0, NULL, 0 }
+	};
+
+	int i;
+	for(i = 0; signals[i].sig != 0; i++)
+		if(signal(signals[i].sig,
+					signals[i].ignore ? SIG_IGN : sig_handler) < 0)
+			fprintf(stderr,"Setting handler for %s: %s\n", signals[i].name,
+					strerror(errno));
+}
+
+static void leave(int sig)
 {
 	fprintf(stderr,"qemu exited: %s quits\n", vdeqname);
+	cleanup();
 	exit(0);
 }
 
-
 static int checkver(char *prog)
-{ 
+{
 	char *newargv[3];
 	int fd[2];
 	int f,len,version=0;
@@ -148,7 +162,10 @@ static int checkver(char *prog)
 	newargv[1]="-h";
 	newargv[2]=0;
 	buf[256]=0;
-	pipe(fd);
+	if (pipe(fd) < 0) {
+	  perror("pipe");
+	  exit(1);
+	}
 	if ((f=fork()) > 0) {
 		int status;
 		close(fd[1]);
@@ -174,19 +191,20 @@ static int checkver(char *prog)
 		if (execvp(prog,newargv) < 0) {
 			exit(1);
 		}
-	} 
+	}
 	return version;
-}   
+}
 
-static char *parsevdearg(char *arg,char **sock,int fd)
+static char *parsevdearg(char *arg,char **sock,int *pport, int fd)
 {
 	char newarg[128];
 	int vlan=0;
 	*sock=VDESTDSOCK;
+	*pport=0;
 	while (*arg==',') arg++;
 	if (strncmp(arg,"vlan=",5)==0) {
-    vlan=atoi(arg+5);
-		while (*arg != 0 && *arg != ',') 
+		vlan=atoi(arg+5);
+		while (*arg != 0 && *arg != ',')
 			arg++;
 	}
 	while (*arg==',') arg++;
@@ -195,11 +213,11 @@ static char *parsevdearg(char *arg,char **sock,int fd)
 		if (*arg=='\"') {
 			arg++;
 			*sock=arg;
-			while (*arg != 0 && *arg != '\"') 
+			while (*arg != 0 && *arg != '\"')
 				arg++;
 		} else {
 			*sock=arg;
-			while (*arg != 0 && *arg != ',') 
+			while (*arg != 0 && *arg != ',')
 				arg++;
 		}
 		if (*arg != 0) {
@@ -207,92 +225,108 @@ static char *parsevdearg(char *arg,char **sock,int fd)
 		}
 	}
 	while (*arg==',') arg++;
+	if (strncmp(arg,"port=",5)==0) {
+		*pport=atoi(arg+5);
+		while (*arg != 0 && *arg != ',')
+			arg++;
+	}
+	while (*arg==',') arg++;
+
 	snprintf(newarg,128,"tap,vlan=%d,fd=%d%s%s",vlan,fd,(*arg == 0)?"":",",arg);
 	return strdup(newarg);
 }
 
 int main(int argc, char **argv)
 {
-  int *fddata;
   char *argsock,**sockname;
-  struct sockaddr_un *dataout,datain;
-  int datainsize;
+	int *ports;
   int result;
-  int group=0;
-  int *connected_fd;
   register ssize_t nx;
-  int args;
   int newargc;
   char **newargv;
   typedef int pair[2];
   pair *sp;
-  register int i;
-  int nb_nics;
-  int oldsyntax=0;
-  int newsyntax=0;
+  register int i,j;
+	int oldsyntax=0;
+	int newsyntax=0;
 	int ver;
+  mode_t mode;
 
   vdeqname=basename(argv[0]);
+	//callerpwd=getpwuid(getuid());
 	/* OLD SYNTAX MGMT */
 	if (strncmp(vdeqname,"vdeo",4) == 0) {
 		oldsyntax=1;
 		if (strcmp(vdeqname,"vdeoq") != 0) {
 			filename=vdeqname+4;
-			args=1;
 		}
 	}
-	else 
-		if (strcmp(vdeqname,"vdeq") != 0 && strncmp(vdeqname,"vde",3)==0) {
-	  filename=vdeqname+3;
-	  args=1;
-  }
-  else if (argc > 1) {
+	else if (strcmp(vdeqname,"vdeq") != 0 && strncmp(vdeqname,"vde",3)==0) {
+		filename=vdeqname+3;
+	}
+	else if (argc > 1) {
 	  filename=argv[1];
-	  args=2;
+		argc--;
+		argv++;
   } else {
 	  usage();
   }
-	if ((ver=checkver(filename)) < 0x800)
+	if ((ver=checkver(filename)) < 0x800) 
 		oldsyntax=1;
 	if (!oldsyntax) {
-		nb_nics=countnewnics(argc-args,argv+args);
+		nb_nics=countnewnics(argc-1,argv+1);
 		if (nb_nics > 0)
 			newsyntax=1;
 	}
-  if ((argc > args && (
-			  strcmp(argv[args],"-h")==0 ||
-			  strcmp(argv[args],"-help")==0 ||
-			  strcmp(argv[args],"--help")==0
+  if ((argc > 1 && (
+			  strcmp(argv[1],"-h")==0 ||
+			  strcmp(argv[1],"-help")==0 ||
+			  strcmp(argv[1],"--help")==0
 			  )) || (
 			  strcmp(filename,"-h")==0 ||
 			  strcmp(filename,"-help")==0 ||
 			  strcmp(filename,"--help")==0
 		  )) {
 	  usage();
-  } else if (argc > args+1 && !newsyntax &&
-		  ((strcmp(argv[args],"-vdesock")==0) ||
-		  (strcmp(argv[args],"-sock")==0) ||
-		  (strcmp(argv[args],"--sock")==0) ||
-		  (strcmp(argv[args],"-unix")==0) ||
-		  (strcmp(argv[args],"-s")==0))
+  } else if (argc > 2 && (
+		  (strcmp(argv[1],"-vdesock")==0) ||
+		  (strcmp(argv[1],"-sock")==0) ||
+		  (strcmp(argv[1],"-unix")==0) ||
+		  (strcmp(argv[1],"-s")==0))
 	    ){
-	  argsock=argv[args+1];
-	  args+=2;
+	  argsock=argv[2];
+	  argv+=2;
+	  argc-=2;
   } else
 	  argsock=NULL;
 
+    if (argc > 2 && (
+		  (strcmp(argv[1],"--mod")==0) ||
+		  (strcmp(argv[1],"-m")==0))
+	    ){
+	sscanf(argv[2],"%o",&mode);
+	argv+=2;
+	argc-=2;
+    }
+
 	if (!newsyntax) {
 		if (argsock == NULL)
 			nb_nics=1;
 		else
 			nb_nics=countnics(argsock);
-		if (!oldsyntax && nb_nics > 1) 
+		if (!oldsyntax && nb_nics > 1)
 			fprintf(stderr,
 					"Warning: all the vde connections will be connected to one net interface\n"
 					"         to configure several interface use the new syntax -net vde\n");
 	}
-  if ((sp= (pair *) malloc(nb_nics * 2 * sizeof (int)))<0) {
-	  perror("malloc nics");
+
+	if ((sp= (pair *) malloc(nb_nics * 2 * sizeof (int)))<0) {
+		perror("malloc nics");
+		exit(1);
+	}
+
+	if ((conn=(VDECONN **) calloc (nb_nics,sizeof(VDECONN *))) <0) {
+	  perror("calloc conn");
 	  exit(1);
   }
 
@@ -303,8 +337,12 @@ int main(int argc, char **argv)
 		}
   }
 
-  if ((sockname= (char **) malloc(sizeof(char *) * nb_nics))<0) {
-	  perror("malloc sockname");
+	if ((sockname= (char **) malloc(sizeof(char *) * nb_nics))<0) {
+		perror("malloc sockname");
+		exit(1);
+	}
+  if ((ports= (int *) calloc(nb_nics, sizeof(int)))<0) {
+	  perror("malloc ports");
 	  exit(1);
   }
 
@@ -314,20 +352,20 @@ int main(int argc, char **argv)
 		int vdeint;
 		newargv=argv;
 		newargc=argc;
-		for (i=0,netflag=0,vdeint=0;i<argc;i++) {
-		 if (strcmp(argv[i],"-net")==0)
-			 netflag=1;
-		 else {
-			 if (netflag && strncmp(argv[i],"vde",3) == 0)
-			 {
-				 argv[i]=parsevdearg(argv[i]+3,&sockname[vdeint],sp[vdeint][0]);
-				 vdeint++;
-			 }
-			 netflag=0;
-		 }
+		for (i=1,netflag=0,vdeint=0;i<argc;i++) {
+			if (strcmp(argv[i],"-net")==0)
+				netflag=1;
+			else {
+				if (netflag && strncmp(argv[i],"vde",3) == 0)
+				{
+					argv[i]=parsevdearg(argv[i]+3,&sockname[vdeint],&ports[vdeint],sp[vdeint][0]);
+					vdeint++;
+				}
+				netflag=0;
+			}
 		}
 	} else
-	{
+  {
 		if (argsock==NULL)
 			sockname[0]=VDESTDSOCK;
 		else
@@ -349,7 +387,7 @@ int main(int argc, char **argv)
 				printf("as %s\n",argsock);
 				for (i=0; i<nb_nics; i++)
 				printf("%d -> %s\n",i,sockname[i]); */
-		newargc=argc+1+2+(2*nb_nics)-args;
+		newargc=argc+2+(2*nb_nics);
 		if ((newargv=(char **) malloc ((newargc+1)* sizeof(char *))) <0) {
 			perror("malloc");
 			exit(1);
@@ -379,35 +417,20 @@ int main(int argc, char **argv)
 			newargv[2*nb_nics+1]="-net";
 			newargv[2*nb_nics+2]="nic";
 		}
-		for (i=(2*nb_nics)+3;args<argc;i++,args++) newargv[i]=argv[args];
+		for (i=(2*nb_nics)+3,j=1;j<argc;i++,j++) newargv[i]=argv[j];
 
 		newargv[i]=0;
 	}
 
-  if ((fddata= (int *) malloc(sizeof(int) * nb_nics))<0) {
-	  perror("malloc fddata");
-	  exit(1);
-  }
-  if ((connected_fd= (int *) malloc(sizeof(int) * nb_nics))<0) {
-	  perror("malloc connected_fd");
-	  exit(1);
-  }
-  if ((dataout= (struct sockaddr_un *) malloc(sizeof(struct sockaddr_un) * nb_nics))<0) {
-	  perror("malloc dataout");
-	  exit(1);
-  }
-
   if ((pollv= (struct pollfd *) malloc(sizeof(struct pollfd) * 2 * nb_nics))<0) {
-	  perror("malloc fddata");
+		perror("malloc pollfd");
 	  exit(1);
   }
+	setsighandlers();
   for (i=0; i<nb_nics; i++) {
-	  if((fddata[i] = socket(AF_UNIX, SOCK_DGRAM, 0)) < 0){
-		  perror("socket");
-		  exit(1);
-	  }
-	  connected_fd[i]=send_fd(sockname[i], fddata[i], &(dataout[i]), i, group);
-	  pollv[2*i+1].fd=fddata[i];
+		struct vde_open_args vdearg={ports[i],NULL,mode};
+		conn[i]=vde_open(sockname[i],"vdeqemu",&vdearg);
+	  pollv[2*i+1].fd=vde_datafd(conn[i]);
 	  pollv[2*i].fd=sp[i][1];
 	  pollv[2*i].events= pollv[2*i+1].events=POLLIN|POLLHUP;
   }
@@ -418,31 +441,50 @@ int main(int argc, char **argv)
 	  for (i=0; i<nb_nics; i++) 
 		  close(sp[i][0]);
 	  for(;;) {
-		  result=poll(pollv,2*nb_nics,-1);
-		  for (i=0; i<nb_nics; i++) {
-			  if (pollv[2*i].revents & POLLHUP || pollv[2*i+1].revents & POLLHUP)
-				  break;
-			  if (pollv[2*i].revents & POLLIN) {
-				  nx=read(sp[i][1],bufin,sizeof(bufin));
-				  //fprintf(stderr,"RX from qemu %d\n",nx);
-				  //send(connected_fd,bufin,nx,0);
-				  sendto(fddata[i],bufin,nx,0,(struct sockaddr *) &(dataout[i]), sizeof(struct sockaddr_un));
-			  }
-			  if (pollv[2*i+1].revents & POLLIN) {
-				  datainsize=sizeof(datain);
-				  nx=recvfrom(fddata[i],bufin,BUFSIZE,0,(struct sockaddr *) &datain, &datainsize);
-				  //fprintf(stderr,"TX to qemu %d\n",nx);
-				  write(sp[i][1],bufin,nx);
-			  }
-		  }
+	    if ((result=poll(pollv,2*nb_nics,-1)) < 0) {
+	      perror("poll");
+	      cleanup();
+	      exit(1);
+	    }
+	    for (i=0; i<nb_nics; i++) {
+	      if (pollv[2*i].revents & POLLHUP || pollv[2*i+1].revents & POLLHUP)
+		break;
+	      if (pollv[2*i].revents & POLLIN) {
+		if ((nx=read(sp[i][1],bufin,sizeof(bufin))) < 0) {
+		  perror("read");
+		  cleanup();
+		  exit(1);
+		}
+		//fprintf(stderr,"RX from qemu %d\n",nx);
+		if (vde_send(conn[i],bufin,nx,0) < 0) {
+		  perror("sendto");
+		  cleanup();
+		  exit(1);
+		}
+	      }
+	      if (pollv[2*i+1].revents & POLLIN) {
+		if ((nx=vde_recv(conn[i],bufin,BUFSIZE,0)) < 0) {
+		  perror("recvfrom");
+		  cleanup();
+		  exit(1);
+		}
+		//fprintf(stderr,"TX to qemu %d\n",nx);
+		if (write(sp[i][1],bufin,nx) < 0) {
+		  perror("write");
+		  cleanup();
+		  exit(1);
+		}
+	      }
+	    }
 	  }
   } else {
 	  for (i=0; i<nb_nics; i++) {
 		  close(sp[i][1]);
-		  close(fddata[i]);
-		  close(connected_fd[i]);
+		  close(vde_datafd(conn[i]));
+		  close(vde_ctlfd(conn[i]));
 	  }
 	  execvp(filename,newargv);
   }  
+	cleanup();
   return(0);
 }
