@@ -45,6 +45,7 @@
 #define WANT_SYS_IOCTL_H
 #include <config.h>
 #include <slirp.h>
+#include "tcp2unix.h"
 
 /* patchable/settable parameters for tcp */
 int 	tcp_mssdflt = TCP_MSS;
@@ -397,10 +398,27 @@ int tcp_fconnect(so)
      struct socket *so;
 {
   int ret=0;
+	char *path;
   
   DEBUG_CALL("tcp_fconnect");
   DEBUG_ARG("so = %lx", (long )so);
 
+	//fprintf(stderr,"fconnect %d %s\n",ntohs(so->so_fport), inet_ntoa(so->so_faddr));
+	if (__builtin_expect(tcp2unix_check,0) &&
+			(so->so_faddr.s_addr & htonl(0xffffff00)) == special_addr.s_addr &&
+			(ntohl(so->so_faddr.s_addr) & 0xff) == CTL_ALIAS &&
+			(path=tcp2unix_search(ntohs(so->so_fport)))!=NULL ) {
+		if ( (ret=so->s=socket(AF_FILE,SOCK_STREAM,0)) >= 0) {
+			int opt, s=so->s;
+			struct sockaddr_un addr;
+
+			fd_nonblock(s);
+			addr.sun_family = AF_FILE;
+			strncpy(addr.sun_path,path,UNIX_PATH_MAX);
+			ret = connect(s,(struct sockaddr *)&addr,sizeof (addr));
+			soisfconnecting(so);
+		}
+	} else
   if( (ret=so->s=socket(AF_INET,SOCK_STREAM,0)) >= 0) {
     int opt, s=so->s;
     struct sockaddr_in addr;
@@ -416,12 +434,12 @@ int tcp_fconnect(so)
       /* It's an alias */
       switch(ntohl(so->so_faddr.s_addr) & 0xff) {
       case CTL_DNS:
-	addr.sin_addr = dns_addr;
-	break;
+				addr.sin_addr = dns_addr;
+				break;
       case CTL_ALIAS:
       default:
-	addr.sin_addr = loopback_addr;
-	break;
+				addr.sin_addr = loopback_addr;
+				break;
       }
     } else
       addr.sin_addr = so->so_faddr;
@@ -461,7 +479,7 @@ tcp_connect(inso)
 {
 	struct socket *so;
 	struct sockaddr_in addr;
-	int addrlen = sizeof(struct sockaddr_in);
+	socklen_t addrlen = sizeof(struct sockaddr_in);
 	struct tcpcb *tp;
 	int s, opt;
 
@@ -513,6 +531,7 @@ tcp_connect(inso)
 		so->so_state = SS_NOFDREF; /* Don't select it yet, even though we have an FD */
 					   /* if it's not FACCEPTONCE, it's already NOFDREF */
 	}
+	/*fprintf(stderr,"S=%d\n",s);*/
 	so->s = s;
 	
 	so->so_iptos = tcp_tos(so);
@@ -656,7 +675,7 @@ tcp_emu(so, m)
 		{
 			struct socket *tmpso;
 			struct sockaddr_in addr;
-			int addrlen = sizeof(struct sockaddr_in);
+			socklen_t addrlen = sizeof(struct sockaddr_in);
 			struct sbuf *so_rcv = &so->so_rcv;
 			
 			memcpy(so_rcv->sb_wptr, m->m_data, m->m_len);
@@ -983,8 +1002,8 @@ do_prompt:
 			}
 			return 0;
 		}
-#endif		
-        case EMU_FTP: /* ftp */
+#endif
+	 case EMU_FTP: /* ftp */
 		*(m->m_data+m->m_len) = 0; /* NULL terminate for strstr */
 		if ((bptr = (char *)strstr(m->m_data, "ORT")) != NULL) {
 			/*
